#!/usr/bin/env python

from typing import Dict
from ase import Atoms
from ase.data.pubchem import pubchem_atoms_search
from ase.calculators.gaussian import Gaussian

# TODO: add class to module
class RIx:
    _format: str = '7.3f'
    def __init__(self, A: Dict[int, Atoms], B: Dict[int, Atoms]) -> None:
        self.A = A
        self.B = B
    
    @property
    def I_A(self):
        atoms_plus = self.A[1]
        atoms_zero = self.A[0]
        e_plus = atoms_plus.get_potential_energy()
        e_zero = atoms_zero.get_potential_energy()
        I = e_plus - e_zero
        return I

    @property
    def I_B(self):
        atoms_plus = self.B[1]
        atoms_zero = self.B[0]
        e_plus = atoms_plus.get_potential_energy()
        e_zero = atoms_zero.get_potential_energy()
        I = e_plus - e_zero
        return I
    
    @property
    def A_A(self):
        atoms_minus = self.A[-1]
        atoms_zero = self.A[0]
        e_minus = atoms_minus.get_potential_energy()
        e_zero = atoms_zero.get_potential_energy()
        A = e_zero - e_minus
        return A

    @property
    def A_B(self):
        atoms_minus = self.B[-1]
        atoms_zero = self.B[0]
        e_minus = atoms_minus.get_potential_energy()
        e_zero = atoms_zero.get_potential_energy()
        A = e_zero - e_minus
        return A

    @property
    def nu_A(self):
        return self.I_A - self.A_A

    @property
    def nu_B(self):
        return self.I_B - self.A_B

    @property
    def mu_A(self):
        return -(self.I_A + self.A_A) / 2

    @property
    def mu_B(self):
        return -(self.I_B + self.A_B) / 2

    @property
    def dNA_ele(self):
        return -(self.A_B - self.I_A) / (2*(self.nu_B + self.nu_A))
    
    @property
    def dNB_ele(self):
        return -self.dNA_nuc
    
    @property
    def dNA_nuc(self):
        return (self.A_A - self.I_B) / (2*(self.nu_B + self.nu_A))
    
    @property
    def dNB_nuc(self):
        return -self.dNA_ele

    @property
    def dNA(self):
        return self.dNA_ele + self.dNA_nuc

    @property
    def dNB(self):
        return self.dNB_ele + self.dNB_nuc

    def str_A(self):
        text = f' {self.I_A:{self._format}} | {self.A_A:{self._format}} | '
        text += f'{self.mu_A:{self._format}} | {self.nu_A:{self._format}} | '
        text += f'{self.dNA:{self._format}} | '
        text += f'{self.dNA_ele:{self._format}} | {self.dNA_nuc:{self._format}}'
        return text

    def str_B(self):
        text = f' {self.I_B:{self._format}} | {self.A_B:{self._format}} | '
        text += f'{self.mu_B:{self._format}} | {self.nu_B:{self._format}}'
        return text

    def __str__(self):
        text = '   I_A   |   A_A   |   μ_A   |   η_A   |   ΔNA   | ΔNA_ele | ΔNA_nuc \n'
        text += self.str_A() + '\n'
        text = '   I_B   |   A_B   |   μ_B   |   η_B   |\n'
        text += self.str_B()
        return text


def main():
    info = dict()
    # list of molecules to calculate
    molecule_A = ['ethene', '1-propene', '(E)-but-2-ene', '(Z)-but-2-ene', '2-methylprop-1-ene', '2-methylbut-2-ene', '2,3-dimethylbut-2-ene']
    molecule_B = ['Br2']
    # Gaussian parameters
    # TODO: make parameters function arguments
    mem = '1GB'
    method = 'M06-2X'
    basis = '6-311G(d,p)'
    scf = 'qc'
    # iterate over molecules
    for formula in molecule_A + molecule_B:
        # get atoms structure from PubChem
        atoms = pubchem_atoms_search(formula)
        # add cell and vacuum
        atoms.center(vacuum=5)

        info[formula] = dict()
        # iterate over charges
        for charge in [-1, 0, 1]:
            # copy Atoms object
            atoms_charge = atoms.copy()
            # calculation name
            name = f'{formula}_{charge}'
            # initialize calculator
            calc = Gaussian(mem=mem,
                            chk=f'{name}.chk',
                            method=method,
                            basis=basis,
                            scf=scf)
            atoms_charge.set_calculator(calc)
            # perform calculation
            atoms_charge.get_potential_energy()
            # save Atoms object
            info[formula][charge] = atoms_charge

    # calculate reaction indices
    maxlen = max(len(formula) for formula in molecule_A)
    text = f'{"molecule":>{maxlen}}|   I_A   |   A_A   |   μ_A   |   η_A   |   ΔNA   | ΔNA_ele | ΔNA_nuc\n'
    for formula in molecule_A:
        rix = RIx(A=info[formula], B=info[molecule_B[0]])
        text += f'{formula:<{maxlen}}|'
        text += rix.str_A() + '\n'
    text += f'{molecule_B[0]:<{maxlen}}|'
    text += rix.str_B()
    # write results to file
    with open('RIx.txt', 'w') as f:
        f.write(text)
    # print results
    print(text)

if __name__=='__main__':
    main()